MODULE MBUFR


!******************************************************************************
!*                               MODULO MBUFR                                 *
!*                                                                            *
!*                                                                            *
!*     MODULO GENÉRICO PARA GRAVACAO/GRAVACAO DE ARQUIVOS EM FORMATO BUFR     *
!*																			  *
!*               CENTRO DE PREVISAO DE TEMPO E ESTTUDOS CLIMATICOS            * 
!*              GRUPO DE PRE-PROCESSAMENTO  DE DADOS METEOROLOGICOS           *
!*                                											  *
!*                              (CPTEC-INPE-MCT)                              *
!*																			  *
!******************************************************************************
! 
! 1 - DESCRICAO GERAL 
! 
!   Este é um módulo, escrito em FORTRAN 90, eh compativel com os sistemas 
!   WINDOWS, LINUX e UNIX e proporciona 3 rotinas publicas que podem ser chamadas
!   de dentro de um programa principas para gravar dados em formato BUFR da WMO
!
!   As rotinas publicas fornecidas por este modulo sao:
!
!    a) OPEN_MBUFR	( un, filename, centre) 
!    b) SAVE_MBUFR 	( un,d,ndesc,v,iconf,nsubset,Btype,Bsubtype,Hobs,idObs)
!    c) CLOSE_MBUFR ( un)
!
!    A SUBROUTINA OPEN_MBUFR Abre um arquivo BUFR de acesso aleatorio, para
!    gravacao dos dados BUFR. Tambem inicializa as estruturas de dados internas
!    deste modulo, incluindo a tabela de descritores BUFR que eh lido do arquivo
!    "BufrTableB_mbufr.txt".  
!     
!    A SUBROTINA SAVE_MBUFR Grave uma mensagem BUFR em um arquivo aberto por
!    OPEN_MBUFR. 
! 
!    A SUBROUTINA  CLOSE_MBUFR(UN)  Fecha um arquivo BUFR  aberto por OPEN_MBUFR
!    e imprime na tela algumas informacoes relativas ao  processo de gravacao
!
!
! 2 - VARIAVEIS DA INTERFACE 
! 
!     -------------------------------------------------------------------------
!     |  un       | Inteiro que designa a unidade Lógica para gravacao, 
!     |           |  Usado em OPEN_MBUFR, SAVE_MBUFR e CLOSE_MBUFR
!     |-----------+-------------------------------------------------------------
!     |  filename | Nome do arquivo BUFR a ser aberto (OPEN_MBUFR) 
!     |-----------+-------------------------------------------------------------
!     |  centre   | Codigo de Centro gerador da mensagem, common code table C-1
!     |           | No caso do CPTEC centre=46
!     |-----------+-------------------------------------------------------------
!     |  d        | Vetor do tipo "descbufr", d(1:ndesc) atraves do qual eh 
!     |           | passado o codigo de cada observacao conforme TABELAS BUFR. 
!     |           | Este código BUFR eh composto por 3 partes (F,X, Y ), que devem
!     |           | ser atribuidos da seguinte  forma:
!     |           |      D(i)%F=Valor de F
!     |           |      D(i)%X=Valor de X
!     |           |      D(i)%Y=Valor de Y
!     |-----------+--------------------------------------------------------------
!     | ndesc     | Numero de descritores em d
!     |-----------+--------------------------------------------------------------
!     | v         | Array real de dados observacionais  v(1:nsubset,1:ndesc) 
!     |-----------+--------------------------------------------------------------
!     | nsubsets  | Numero de subsets, ou numero de observacoes em v
!     |-----------+--------------------------------------------------------------
!     |	BType     | Tipo da Observacao BUFR, conforme tabelas WMO 
!     |           | Exemplo: Para o caso de observacoes de superficie BType=0
!     |-----------+--------------------------------------------------------------
!     | Bsubtype  | Subtipo da Observcao, conforme tabela local do centro
!     |-----------+--------------------------------------------------------------
!     | HObs      | Variavel do tipo MBUFR_Date, atraves da qual e passado 
!     |           | a data e hora da observacao  a ser gravado na secao 1 BUFR.
!     |           |
!     |           |  Ex.: Hobs%year = ano , Hobs%month=mes, etc.
!     |-----------+--------------------------------------------------------------
!     |  IdObs    | Variavel do tipo MBUFR_ID, ataves do qual e passado os
!     |           | dados de posicao da estacao meteorlogica e  numero da estacao,
!     |           | a serem incluidos na secao opcional 2 do BUFR
!     |           |  
!     |           | Ex.:  idObs%lat = Latitude, idIbs%Lon = Longitude, etc.
!     |           |
!     |           | Nota: Se idObs for omitido na chamada de SAVE_MBUFR,a secao 2
!     |           | nao sera gravada 
!     ----------------------------------------------------------------------------
!
! 3 - ROTINAS PRIVATIVAS (Acessiveis apenas internamente)
!
!    PUT_OCTETS,
!    SAVESEC0, SAVESEC1, SAVESEC2, SAVESEC3, SAVESEC4, SAVESAC5
!    
!    INIT_TABB
!    
! 4 - VARIVAEIS GLOBAIS PRIVATIVAS (Visiveis a todas as rotinas internas do 
!     modulo, porem nao visiveis externamente)
!
!     currentRG, verloctab, vermastertab, NMSG, subname, centre_mbufr 
!
! 5 - ARRAYS GLOBAIS PRIVATIVOS (Visiveis a todas as rotinas internas do 
!     modulo, porem nao visiveis externamente)
!
!     TABB (Tabela de descritores BUFR)
!-------------------------------------------------------------------------------
!
! 6 - HISTORICO
! 
!     VERSAO 0.1 -  SERGIO H. S. FERREIRA  Iniciado 17/07/2004 
!                   Extruturacao básica do modulo e reaproveitamento de
!                   da subrotina GRAVA_BUFR do programa das PCDs
!
!     Versao 0.2 -  SERGIO H. S. FERREIRA 
!                   Desenvolvimento da subrotina PUT_OCTETS
!
!     Versao 0.3 -  SERGIO H.S FERREIRA 26/07/2003
!                   Primeira Versao para teste de gravacao do dados de SATOB 
!                   do DSA, a partir de chamadas de dentro do programa CDVENSAT 
!
!     
!     Nota: Esta eh  um modulo experimental
!    
!*****************************************************************************

   IMPLICIT NONE    ! Todas as variaveis serao declaradas
   
   PRIVATE          ! Todas as Variaveis, estrutuiras e subrotinas 
                    ! serao, em principio, privadas

 !{ Declara como PUBLICAS as seguintes  subrotinas 
 
   PUBLIC OPEN_MBUFR
   PUBLIC SAVE_MBUFR
   PUBLIC CLOSE_MBUFR
!}

!{ Declara como PUBLICAS os seguintes tipos de estruturas

   PUBLIC descbufr	     ! Estrutura para armazenar descritores BUFR 	  
   PUBLIC MBUFR_Id		 ! Estrutura p/ armazenar identificacao de estacao 
   PUBLIC MBUFR_Date	 ! Estrutura para armazenar data e  hora
  
 !}


!{ Definicao  dos tipos PUBLICOS de estruturas  do modulo  

  
    type MBUFR_Id				
      character(len=20):: COD          ! Codigo da estacao
      character(len=20)::NAME         ! Nome da Localidade ou estacao
      Real:: Lat,lon				  ! Coordenadas Horizontais da estacao
      real::	Ha				      ! Altitude da estacao ou localidade
	  real :: conf                    ! Confiabilidade                   
    end	type MBUFR_Id
 

    type MBUFR_Date				
      Integer:: MES,DIA,ANO,HORA,MIN            
    end type MBUFR_date  
 	   
 
   type descbufr				  
     integer :: F
     integer :: X
     integer :: y
   end type
  


  !{ Definicao  estrutura PRIVATIVA para armazenar tabelas BUFR 

   type bufrtabletype     
     real :: scale	    ! Fator de Escala 
     real :: refv	    ! Valor de Referência 
     integer :: nbits   ! Número de Bits
   end type
  !}




!{ ALOCACAO DAS VARIAVEIS GLOBAIS E PRIVATIVAS DO MODULO
  
	   integer::currentRG           ! Posicao Corrente no arquivo BUFR
	   integer::centre_mbufr        ! Código do Centro que gerou o BUFR
	   integer::verloctab	        ! Versao da tabela local 
	   integer::vermastertab        ! Versao da tabela Master
	   character(len=5)::STATION_NUMBER 
	   integer :: NMSG              ! Número de mensagens
       character (len=255)::subname ! Guarda nome da subrotina para fins de DEBUG
 !}

!{ ALOCACAO DA TABELA B BUFR

	   type(bufrtabletype),dimension(0:0,0:35,0:200)::tabb 
!}

!----------------------------------------------------------------------------------


CONTAINS






















!***********************************************************
! SUBROTINA PUBLICA OPEN_MBUFR
!
! ABRI UM ARQUIVO BUFR e INICIALIZA AS VARIAVEIS DE
! CONFIGURACAO GERAL DO MÓDULO MBUFR
!-----------------------------------------------------------


SUBROUTINE OPEN_MBUFR(UN,filename,centre)

   integer,intent(in)::un			         !Unidade Lógica para gravacao
   character(len=*),intent(in)::filename	 !Nome do arquivo
   integer,intent(in)::centre                !Código do Centro Gerados
   integer::uni								 !Variavel auxiliar de UN
 
   uni=un
   VerMasterTab=2
   VerLocTab=1
   centre_mbufr=centre
   currentRG=0
   NMSG=0
 
   call INIT_TABB(Uni)                        ! Carrega tabela BUFR  em TABB

   open(uni,file=filename,STATUS='unknown',FORM='UNFORMATTED',access='DIRECT',recl=1) 
   print *," Arquivo BUFR=",trim(filename)
   print *, " "


 end subroutine OPEN_MBUFR


















! ---------------------------------------------------------------------
!
!                       SUBROUTINE PUBLICA SAVE_MBUFR
!
!  ESTA ROTINA E GRAVA UMA MENSAGEM BUFR COMPOSTA POR 6 SECOES (de 0 a 5) 
!  
!  Primeiramente e feito um salto de 8 bytes no arquivo, deixando espaco
!  para a secao 0, que sera gravada no final 
! 
!  Em seguida sao gravados as secoes de 1 a 5. que possuem os seguintes 
!  tamanhos
!    secao 1 - 18 Bytes
!    secao 2 - 52 Bytes (Optional)
!    secao 3 - Variavel (Depende do Número de descritores )
!    secao 4 - Variavel (Depende das variaveis e do numero de  subsets)
!    secao 5 - 4 bytes "7777"
!
!  Apos a gravacao destas secoes eh feito o calculo do tamanho da mensagem
!  que eh gravado na secao 0
!
!  Alguns testes de tamanho tambem sao realidados
!  
! ---------------------------------------------------------------------


SUBROUTINE save_mbufr(un,d,ndesc,v,iconf,nsubset,IBtype,IBsubtype,Hobs,idObs)
 
 !{ VARIAVEIS DA INTERFACE EXTERNA 
   integer,intent(in)::UN            
   type(descbufr),dimension(:),intent(in)::D                                    
   integer,intent(in)::ndesc         
   real,dimension(:,:),intent(in)::v                                    
   integer,dimension(:,:),intent(in)::iconf 
   integer,intent(in)::nsubset       
   Integer,intent(in):: IBtype       
   Integer, intent(in)::IBsubtype     
   type(MBUFR_Date):: HObs           
   type(MBUFR_Id),optional:: IdObs   
 !}
  	
 !{ DEFINICOES DE VARIAVEIS LOCAIS

 logical::sec2opt !- Se verdadeiro a secao 2 sera gravada
 INTEGER Tam_BUFR !-Tamanho total do arquivo BUFR (grav. na secao 1)
 INTEGER Tam_sec2 !-Numero de bytes da secao 2
 INTEGER Tam_sec3 !-Numero de bytes da secao 3
 INTEGER Tam_sec4 !-Numero de bytes da secao 4
 INTEGER BTYPE,BSUBTYPE


 INTEGER UNI                ! Numero da Unidade Lógica de Gravacao 
 
 INTEGER :: RG0,RGF,NRG     ! RG de Inicio, Fim e Numero de RG de uma 
                            ! Mensagem BUFR
 INTEGER :: RG2             ! Registro de Inicio da Secao 2

  UNI=UN
  BTYPE=IBTYPE
  BSUBTYPE=IBSUBTYPE
 
 
 ! Se IdObs for fornecido entao a secao 2 será utilizada
 ! { 
   if (present(idObs))  then
     sec2opt=.true.
   else 
     sec2opt=.false.
   end if
 !}




  RG0=currentRG                !Guarda o Registro da fim da mensagem anterior         				   					
  
  !call savesec0(un,Tam_Bufr)  
  currentRG=RG0+8              !Salta para fim da secao 0 

!---------------------------------------------------------------------------
! { Salva Secoes de 1 a 5

   call savesec1(un,sec2opt,btype,bsubtype,Hobs)   
 
   !{Salva a Secao 2 se savesec2_mbufr=.true.    
   
     if (sec2opt) then
	   RG2=currentRg           ! Guarda a Posicao da secao2  	   
	   currentRG=currentRG+52  ! Salta os 52 bytes da secao2 
	                           ! para gravar posteriormente
	                            
	   tam_sec2=52
	 
	 else
	   tam_sec2=0

	 end if  
  !}
   
    call savesec3(un,d,ndesc,nsubset,Tam_Sec3)           
    call savesec4(un,d,ndesc,v,nsubset,iconf,Tam_Sec4)  
    call savesec5(un)                            

!} ----------------------------------------------------------------------------


  RGF=CurrentRG ! Guarda a posicao do registro do fim da mensagem


!{ Calcula o Tamanho do mensagem e verifica se o tamanho está coerente com
!  a posicao do registro corrente (currentRG)
  
  Tam_bufr=8+18+tam_sec2+Tam_sec3+Tam_sec4+4
  NRG=RGF-RG0
  if (Tam_bufr/=NRG) then 
    print *, "Erro em SAVE_MBUFR"
	print *, ""
	close(un)
	stop
  end if     
! }


!{ Grava a secao 0 da mensagem corrente e  reposciona o curentRG para gravacao
!  da proxima mensagem

   currentRG=RG0                   !Reposiciona registro do inicio da mensagem
   call savesec0(un,Tam_Bufr)      !Salva a secao 0 

   !{ Se a secao 2 estiver inclusa, 
   ! Posiciona regitro no inicio da secao 2 e grava a secao 2

   if (sec2opt) then  
     currentRG=RG2                 
	 call savesec2(un,Hobs,IdObs,Tam_Bufr,btype,bsubtype)
   end if
   !}

   currentRG=RGF                   !Reposiciona registro no fim da mensagem 
   NMSG=NMSG+1					   !Incrementa o numero de mensagens gravadas

!}
 
 
 
 END SUBROUTINE




















! SUBROTINA PUBLICA CLOSE_MBUFR - FECHA UM ARQUIVO ABERTO POR OPEM_MBUFR

SUBROUTINE CLOSE_MBUFR(UN)
  integer,intent(in):: UN
  
  close (un)
  Print *, NMSG," Mensagens gravadas"
  print *, currentRg," Bytes gravados"
  print *,""


END SUBROUTINE 

















 !*****************************************************************************
 !                   SUBROUTINE PRIVADA PUT_OCTETS                            *
 !																			  *
 ! FUNCAO: 																	  *
 !  COPIA UM VETOR DE VALORES INTEIROS COM DIFERENTES NUMEROS DE BITS PARA UM *
 !  VETOR DE ELEMENTOS DE 8 BITS (OCTETO)                                     *
 !																			  *
 ! UTILIZACAO:																  *
 !																			  *
 !  Dentro da arquitetura dos computadores,a unidade numérica básica e o Byte *
 !  de 8 bits (OCTETO), de forma que todas as variaveis dentro da memória dos *
 !  computadores, ou gravados em disco, ocupam múltiplos de bytes (8bits).	  *
 !																			  *
 !  Esporádicamente, para fins de compactacao em disco, procede-se o reaprove-*
 !  itamento dos bits de uma variavés, para guadar outras. É o caso dos arqui-*
 !  vos do formato BUFR, que utiliza esta técnica largamente.				  *
 ! 																			  *
 !  Esta subrotina permite que este reaproveitamento seja feito utilizando a  *
 !  seguinte estrutura														  *
 !																			  *
 !	  A_VAL(NA) --> Vetor Inteiro com os NA dados de entrada				  *
 !    A_BITS(NA)--> Vetor Inteiro com o numero de BITS, que deseja-seutilizar *
 !                  para gravar cada elemento de A_VAL						  *
 !    NA        --> Númeor de elementos de A_VAL A_BITS						  *
 !    																		  *
 !    OCT(NOCT)  <-- Vetor Caracter (de 1 Byte (OCTETO)						  *
 !    NOCT       <-- Número de Elementos de OCT								  *
 !	  ERR        <-- Zero se a cópia foi bem sucedida                         *
 !   Utilizando esta estrutura cada elemento de A_VAL é copiado Bit a Bit para* 
 !   OCT, de forma a sempre completar os oito bits de OCT. Os bits que não 	  *
 !   cabem em OCT, sao copiados para o próximo elemento de OCT, de forma que  *
 !   todos os elementos de OCT terá seu bits completados, com excessão do 	  *
 !   último OCT. Este poderá, eventualmente ter seus bits incompletos.        *
 !   Neste caso, a variavel err retornará o valor do número de bits faltantes *
 !   no ultimo octeto														  *
 !   																		  *
 !   IMPORTANTE. 															  *
 ! 																			  *
 !      1- Certifique-se que o último OCT tenha bits completos, fornecendo    *
 !         sempre um conjunto de A_VAL, cuja soma dos bits de todos os        *
 !         elementos sejam múltiplos de 8                                     *
 !																			  *
 !      2- Esta rotina está preparada para copiar elementos de A_VAl que não  *
 !         exceda o tamanho de 64 bits. No caso deste tamanho ser excedido 	  *
 !         poderá ocorrer erro de dimensionamento da matriz local BIT         *
 !																			  *
 !      3- Esta rotina, pressupõe que os bits fornecidos em A_BITS(:), são    *
 !         realmente suficientes para guardar os valores A_VAL(:). Caso seja  *
 !         insuficiente, os bits mais significativos serão cortados. 		  *
 !         Esta é uma parte que precisa ser melhorada. Poderia ser colocado   *
 !         algum teste para verificar e enviar uma mensagem de erro!          *                      
 !          																  *
 !																			  *
 !   EXEMPLO:																  *
 !             Neste Exemplo são fornecidos 5 Valores A_VAL, cada um com res- *
 !             pectivamente 14, 3, 7, 5 e 4 bits.  Neste exemplo esta rotina  *
 !             retorna 5 octetos (OCT), que contem os bits de A_VAL redistri- *
 !             buidos conforme o esquema abaixo                               *              								
 ! 																			  *
 !   A_VAL    11111111111111 000  1010101 00000  1111--+                      *
 !            \---+--/\---+---/\----+---/ \----+---/   |                      *
 !                |       |   		|   	   |	   |					  *
 !            +------+ +------+	 +------+	+------+   |					  *
 !            |      | |      |  |      |   |      |   |					  *
 !  OCT       11111111 11111100  01010101   00000111   10000000               *
 !                                                                            *
 !																			  *
 !            Nete caso, os 5 Elementos de A_VAL são copiados em 5 octetos.   *
 !			  Contudo no Octeto 5, somente 1 bit é utilizado. Os demais 7bits *
 !            ficam com zero. Por isto, neste caso é retornado  err=7         *
 !                                                                            *
 !																			  *
 !*****************************************************************************
 
 subroutine PUT_OCTETS(A_VAL,A_BITS,NA,OCT,NOCt,err)

 integer,dimension(:),intent(inout)::A_VAL	       ! Conjunto de valores de entrada com bits diferentes
 integer,dimension(:),intent(in)::A_BITS           ! Vetor com o numero de bits do conjunto  A_VAL
 integer,intent(in)::NA                            ! Numero de elementos de A_VAL e A_BITS
 character(len=1),dimension(:),intent(out)::OCT    ! Vetor caracter com octetos de saida
 INTEGER,INTENT(OUT)::nOCT                         ! Numero de elementos de OCT
 INTEGER,INTENT(OUT)::err						   ! Indica se o últmo octeto foi completado corretamente.
                                                   !  err =  0     :Todos os bits foram completados
 												   !  err de 1 a 7 :Número de bits que faltaram  
 
 !{ Variaveis locais
 												  
   LOGICAL,dimension(8,NA*10)::BIT  ! Array auxiliar para redistribuicao de bits
   INTEGER:: k,j,i,n,valmax
   byte::V
   character(len=1)::aux
!}


!{ redistribuir os bits da variavel A_VAL em octetos (Vetor BIT)

  K=0
  j=1
  BIT(1:8,J)=0               ! Zera os bits do primeiro octeto
  DO N=1,NA	
  
    !{Verificar se número de bits está dentro do esperado
    	
    if (A_BITS(n)>64) THEN 
	  PRINT *,"Erro em"//TRIM(SUBNAME)//"_copy2oct !"
	  PRINT *, "Possivel erro na especificacao do tamanho da variavel ",N
	  print *, "Numero de bits=",A_BITS(n)
	  stop
	end if
	!}

	!Verificar se o valor fornecido é compatível com
	!o número de bits fornecidos
	!{
	  valmax=0
	  do i=1,A_BITS(N)
	   valmax=valmax+2**(i-1)
	  end do

	  if ((A_VAL(n)<0).or.(A_VAL(n)>valmax)) then
	  	PRINT *,"Erro em "//TRIM(SUBNAME)//"_copy2oct !"
		write(*,500)N,A_VAL(n)
500		FORMAT(	"    Variavel(",I3.3,") =",I11)
		
		IF (SUBNAME=="SAVESEC4") THEN 
		  A_VAL(n)=valmax
		  print *,"   Substituido por valor missing"
		  print *,""
		ELSE
		  STOP
		ENDIF
	  end if

	  
    DO I=1,A_BITS(N)		
      K=K+1
    	 IF (K>8) THEN
	     J=J+1;K=1
	     BIT(1:8,J)=0         ! Zera os bits para armazenar o próximo octeto 
	  END IF
	 BIT(9-K,J)=IBITS(A_VAL(N),A_BITS(N)-i,1)
   END DO
 END DO
 !} Fim da redistribuicao

 NOCT=J


!{ Converter os octetos BIT para valores decimal(V) e para variavel string (oct)
do j=1,noct
  v = 0 
  do i = 1,8
    v = BIT(I,J)*2 ** (i-1) + v    !
  end do
  aux=char(v)
  oct(j)=aux
end do
!} Fim da conversao para string
 
 !err=0 Indica que o ultimo octeto foi completado corretamente
 ! caso contrário, err indica o número de bits que faltaram oara
 ! completar o últmo octeto  
 !{
   err=8-k
  !}

 END SUBROUTINE	PUT_OCTETS























!*******************************************************************************
!*		                        GRAVA A SECAO 0                                *
!*											                                   *
!*******************************************************************************
!   Esta Subrotina grava a Secao 0 de uma mensagem BUFR
!
!
!    As Variáveis de entrada são
!      UN         Unidade Lógica de Gravacao
!      Tam_BUFR   Tamanho da Mensagem até 100000 Bytes 
!
!  
!    Nesta secao são gravados as seguintes informaçoes:
!     a) A palabra "BUFR"  nos  4 primeiros Bytes  
!     b) A Variável Tam_BUFR com 3 Bytes 
!     c) O Número da edicao do código BUFR (1Byte)
!
!    Obsercoes 
! 
!    1 - Note que Tam_BUFR poderia assumir valores de até 8388607, que é o maior
!        inteiro postivo com 24 bits.  Contudo, por convenvesão, uma mensagem
!        BUFR poderá ter no máximo 10000 Bytes. Deta forma, esta subroutina 
!        imprime na tela um aviso quando for ultrapassado  o limite de 10000.  
!        No caso de tentativa de gravar uma mensagem BUFR acima de 8388607 uma 
!        mensagem de erro é mostrado e a gravação é cancelada    
! 
!    2 - Variáveis internas Importantes
!        A()    é um vetor de valores inteiros de 4 bytes
!        B()    é um vetor de  número de bits utilidos para representar cada 
!               valor de A()
!        sec0() é o vetor character que armazena os octetos para gravacao da 
!               secao 0. Este vetor é obtido a partir da copia sequencial 
!               bit-a-bit de cada valor de A(), através da subroutina copy2oct
!------------------------------------------------------------------------------ 
 SUBROUTINE SAVESEC0(UN,Tam_BUFR) 
   
   integer, intent(in)::UN				  ! Unidade Lógica
   integer, intent(in)::Tam_BUFR          ! Tamanho da mensagem BUFR
   
   !{Variáveis Internas 
     character(len=1),dimension(8)::sec0  ! Vetor de octetos	 (1byte)
     integer :: uni,err,noct,i			  ! Outras Variáveis Auxiliares 
     INTEGER,DIMENSION(6)::A,B			  ! Vetor de Valores Inteiros 4 bytes a ser convertidos
   !}
                                             
   uni=un	
  
   SUBNAME="SAVESEC0"   ! Para fins de controle de erros

   !{ Verifica erro no tamanho da mensagem

	if (Tam_BUFR > 3888607) then 
	  
	  print *, "Erro em MBUFR"//SUBNAME
	  print *, "Mensagem muito grande. Tamanho= ", Tam_BUFR
	  print *, "Numero do Mensagem =",NMSG
	  print *, ""
	  close(uni)
	  stop 
    end if
	
	if (Tam_BUFR > 10000 ) THEN
	  print *," Aviso ! Tamanho da Mensagem excedeu limite de 10KBytes "
	  print *," Mensagem =", NMSG
	  print *," Tamanho = ", Tam_BUFR
	  print *,""
	end if

   !}
   
   !{ Prepara para gravar secao 0

	B(1)= 8;A(1)=ichar("B")
	B(2)= 8;A(2)=ichar("U")
	B(3)= 8;A(3)=ichar("F")
	B(4)= 8;A(4)=ichar("R")
    B(5)=24;A(5)=Tam_BUFR   ! Tamanho total do arquivo
    B(6)= 8;A(6)=3          ! BUFR edicao 3 
 

 !{Organiza os dados da secao 0 em octetos (sec0)
 ! Neste caso o numero de octeros noct tem que ser sempre 8

   call PUT_OCTETS(A,B,6,sec0,noct,err)  

   if ((noct/=8).or.(err/=0)) then 
     print *,"Erro na especificacao da secao 0 : numero de octetos =",noct
	 print *,""
	 close(uni)
	 stop
   end if

 !}  
   

 !{Inverte a ordem dos octetos 5,6 e 7 para 7,6 e 5
 
 
 !}

 !{ Grava a secao 0
   do i=1,8
    currentRG=currentRG+1
    write (uni,rec=currentRG) sec0(i)			 
   end do

  !}

 END SUBROUTINE SAVESEC0
   



















!***************************************
!*		     GRAVA A SECAO 1		   *
!***************************************

SUBROUTINE SAVESEC1(UN,sec2opt,btype,bsubtype,HObs)
  integer, intent(in)::UN
  integer, intent(in)::btype
  integer, intent(in)::bsubtype
  type(mbufr_date),intent(in)::Hobs
  logical,intent(in)::sec2opt
  character(len=1),dimension(18):: sec1	          ! Para gravar os 18 bytes da secao 1
  integer :: uni,noct,err,I
  INTEGER,DIMENSION(16)::A,B ! Vetor de Valores e Numero de bits de cada valor

  uni=un
  SUBNAME="SAVESEC1"

 B(1)=24; A(1)=18        !-Contem o tamanho da secao 1 (Bytes de 1 a 3) que é sempre 18 ?
 B(2)=8 ; A(2)=0         !-Contem a versao da BUFR master table  (Byte 4)
                         ! Para meteorologia=0.
 B(3)=16; A(3)=centre_mbufr !-Contem o codigo do  centro que gerou o arquivo  BUFR,
					     ! segundo a code table 0 01 031 (ver manual WMO). 
						 ! No caso, o codigo do INPE (Brazilian Space Agency !?) 
					     ! eh 46  (Bytes de 5 a 6)  
 B(4)=8 ; A(4)=0         !-Update sequence number = 0 (byte 7) 
 B(5)=1 ; A(5)=sec2opt   !-Se este bit=0 entao não há seção 2 (byte 8, bit 0) 
 B(6)=7 ; A(6)=0         ! Estes bits sao reservados  (byte 8, bits de 1 a 7) 			
 B(7)=8 ; A(7)=btype     !-Data category type (BUFR Table A)  (Byte 9)
 					     !  Exemplo: 0 = surface data/land
 B(8)=8	; A(8)=bsubtype  !-Data category sub-type
 					     !  Depende do centro gerador 
 B(9)=8; A(9)=VerMasterTab  !-Versao utilizada da tabela Master/WMO (Byte 11) 
 B(10)=8;A(10)=VerLocTab    !-Versao utilizada da tabela local/INPE (Byte 12)
 
! Um detalhe importante nos proximos bytes (13-18), e o fato de que centro dife-
! rentes estao usando padroes diferentes aqui !!! - A ordem a seguir eh a reco -
! mendada pela WMO, mas o NCEP, por exemplo, poe o ano nos bytes 13 e 14, fazen-
! do um shift nos demais dados e o minuto no byte 18. 
! Devido a estas diferencas, uma boa politica pode ser: nao usar a data que es -
! tiver nesta secao, ou decodificar de acordo com o centro que gerou a mensagem.

 B(11)=8; A(11)= Hobs%ANO - 2000 ! Ano do Seculo (Byte 13)
 B(12)=8; A(12)= Hobs%MES        ! Mes da Observacao (Byte 14)
 B(13)=8; A(13)= Hobs%Dia        ! Dia da Observacao (Byte 15)
 B(14)=8; A(14)= Hobs%HORA  	 ! Hora da Observacao (Byte 16)
 B(15)=8; A(15)= Hobs%MIN	     ! Minuto da Observacao (Byte 17)
 B(16)=8; A(16)= 0  			 ! Reservado para uso local - (Byte 18)

 ! Organiza os dados da secao 1 em octetos (sec1)
 ! Neste caso o numero de octeros noct tem que ser sempre 18
 !{

   call PUT_OCTETS(A,B,16,sec1,noct,err)  

   if ((noct/=18).or.(err/=0)) then 
     print *,"Erro na especificacao da secao 1 : numero de octetos =",noct
	 stop
   end if
 !}


!Grava os 18 byte da secao 1  (SHSF)
!{
  do i=1,18
   currentRG=currentRG+1
   write (uni,rec=currentRG) sec1(i)			 
  end do
!}

END SUBROUTINE SAVESEC1























!******************************************
!* GRAVA A SECAO 2	  Sempre com 52 bytes *
!******************************************

SUBROUTINE SAVESEC2(UN,Hobs,IdObs,Tam_bufr,btype,bsubtype)
 integer, intent(in)::UN
 type(mbufr_date),intent(in)::hobs
 type(mbufr_id),intent(in)::idobs
 character(len=1),dimension(52):: sec2            ! Nesta versao a secao 2 eh uma copia do
 integer,intent(in)::Tam_bufr,btype,bsubtype
 INTEGER(kind=4) rdbmes	    ! LAC
 INTEGER(kind=4) rdbdia	    ! LAC
 INTEGER(kind=4) rdbano	    ! LAC
 INTEGER(kind=4),DIMENSION(8)::rdbtime	 ! LAC
 CHARACTER (12),DIMENSION(3)::RDB_CLOCK  !LAC
 INTEGER,DIMENSION(51)::A,B ! Vetor de Valores e Numero de bits de cada valor
            			                          

 integer :: uni,I,noct,err
 uni=un
 SUBNAME="SAVESEC2"

 B(1)=24 ; A(1)=52       ! Tamanho da secao é sempre 52 (Bytes 1 a 3)
 B(2)=8  ; A(2)=0    	 ! Byte 4 eh reservado (=0)
 B(3)=8  ; A(3)=btype    ! Tipo do dado                 (Byte 5)
 B(4)=8  ; A(4)=bsubtype ! Sub-tipo                     (Byte 6)
 
 !{ Bytes de 7 a 11 
 B(5)=12  ;A(5)=Hobs%ANO     ! Ano 
 B(6)=4   ;A(6)=Hobs%MES	 ! Mes
 B(7)=6   ;A(7)=Hobs%Dia	 ! Dia
 B(8)=5   ;A(8)=Hobs%HORA    ! Hora  
 B(9)=6   ;A(9)=Hobs%MIN     ! Minuto
 B(10)=6  ;A(10)=0           !Segundos 
 B(11)=1  ;A(11)=0			 ! Um Bit zero, provavelmente apenas para
                             ! completar o Byte 11

 !}

 !{ Longitude (Byte 12 a 15)
 B(12)=26 ;A(12)=IdObs%Lon*10**5-(-18000000)  
 B(13)=6  ;A(13)=0
 !}

 !{ Latitude (Bytes de 16 a 19)
 B(14)=25 ;A(14)=(IdObs%Lat*10**5)-(-9000000)
 B(15)=7  ;A(15)=0
 !}

!{ Nome da Estacao (Bytes de 20 a 36) 

 DO I=16,32
   B(I)=8;A(I)=ichar(IdObs%cod(I:I)) 
 END DO							     ! Este é um truque para carregar
                                     ! uma variavel string como inteiros
								

B(33)=16 ; A(33)=Tam_BUFR 	     ! Tamanho Bufr (Bytes 37 e 38)

!{--------------------------------------------------------------------------
!  Agora a hora em que o boletim foi escrito no RDB (faz sentido no CPTEC?)
! --------------------------------------------------------------------------
!  Utiliza data e hora do sistema
!  Bytes de 39 a 41, mais 3 bytes
!

  CALL DATE_AND_TIME(RDB_CLOCK(1), RDB_CLOCK(1), RDB_CLOCK(3), rdbtime)  
  rdbano=rdbtime(1)
  rdbmes=rdbtime(2) 
  rdbdia=rdbtime(3)

  B(34)=6;A(34)=rdbdia
  B(35)=5;A(35)=rdbtime(5) ! Hora
  B(36)=6;A(36)=rdbtime(6) ! Minutos  
  B(37)=6;A(37)=0          ! Segundos
  B(38)=1;A(38)=0          ! Completa o Byte 41

!}

!{ Acrescenta, temporariamente,  3 bytes zeros (Bytes de 42 a 44 
  B(39)=8;A(39)=0
  B(40)=8;A(40)=0  	
  B(41)=8;A(41)=0
!}
  
!{ Byte 45  	
  b(42)=2;a(42)=1  !Primeira Parte da Mensagem
  b(43)=6;a(43)=0
!}
  
!{ Byte 46  
    
   b(44)=2;a(44)=0 ! Número de Correção (No momento deichei fixo. 
   b(45)=6;a(45)=0                ! Futuramente podemos rever isto)
!}

!{ Bytes 47 e 48
   b(46)=8; a(46)=0
   b(47)=8; a(47)=0
!}
 
!{ Byte 49
   b(48)=8;a(48)=idObs%conf ! Coeficiente de Confiabilidade
!}

!{ Bytes de 50 a 52
   b(49)=8;a(49)=0
   b(50)=8;a(50)=0
   b(51)=8;a(51)=0   

 
 ! Organiza os dados da secao 2 em octetos (sec2)
 ! Neste caso o numero de octeros noct tem que ser sempre 52
 !{

   call PUT_OCTETS(A,B,51,sec2,noct,err)  

   if ((noct/=52).or.(err/=0)) then 
     print *,"Erro na especificacao da secao 2 : numero de octetos =",NOCT
	 print *,"err =",err
	 stop
   end if
 !}

 
!Grava os 52 byte da secao 2  (SHSF)
!{
  do i=1,52
   currentRG=currentRG+1
   write (uni,rec=currentRG) sec2(i)			 
  end do
!}


   end subroutine 























!********************
!*  GRAVA A SECAO 3	*
!********************
! TAMANHO DA SECAO 3 : Esta secao possui 7 bytes de identificacao, mais a
! lista de descritores. Cada descritor usa 2 bytes, e como no momento uso
! 17 descritores, assim temos TAMANHO = 7+34 = 41 bytes). Esses valores /
! vao mudar, e o numero de descritores eh definido pela constante Ndescrit
! o Tamanho da secao 3 eh calculado no inicio desta rotina: Tam_sec3



 SUBROUTINE SAVESEC3(UN,D,Ndesc,NSUBSET,tam_sec3) 
  	
 integer, intent(in)::UN
 type(descbufr),dimension(:),intent(in)::D ! Descritores BUFR
 integer,intent(in)::Ndesc                 ! Numero de descritores
 integer,intent(inout)::tam_sec3
 integer,intent(in)::Nsubset
 character(len=1),dimension (1000):: sec3	   ! A identificacao da secao 3 tem 7 bytes
 INTEGER,DIMENSION(8000)::A,B ! Vetor de Valores e Numero de bits de cada valor
 integer :: uni	,ib	,i,err,noct,xx
 uni=un
 SUBNAME="SAVESEC3"

 ! Tamanho da secao 3 corresponde a some de a 7  Bytes que descrevem
 !  a secao  + 2* Nubero total de descritores. 
 !  O Numero total de descritores corresponde ao numero
 !  de descritores das variaveis +1  Replicador +1 De confiabilidade
 

 Tam_sec3 = 7 + 2*(Ndesc+3) !Tamanho secao 3 - descritores
 
 if (mod(Tam_sec3,2)>0 ) Tam_sec3=Tam_sec3+1

B(1)=24;A(1)=Tam_sec3 ! Tamanho da Secao (bites de 1-3) 					  
b(2)=8;a(2)=0         ! byte reservado   (bite 4)
b(3)=16;a(3)=Nsubset  ! Numero de data subsets  (observacoes em cada registro  BUFR) byte
b(4)=1;a(4)=1         ! se 1 Indica dados observacionais
b(5)=1;a(5)=0         ! se 1 Indica dados comprimidos
b(6)=6;a(6)=0         ! Demais bits do octeto é 0  (byte 7)

!A partir do  byte 8 comeca a gravacao dos descritores

ib=6
do i=1,Ndesc
 if (TABB(D(I)%F,D(I)%X,D(I)%Y)%NBITS>0) then 
  ib=ib+1;b(ib)=2;a(ib)=D(I)%F
  ib=ib+1;b(ib)=6;a(ib)=D(I)%X
  ib=ib+1;b(ib)=8;a(ib)=D(i)%Y
 else 
  print *,"Erro ! Descritor BUFR desconhecido"
  write(*,'("Descritor = ",i1,1x,i2,1x,i3,1x)')D(i)%F,D(i)%X,D(i)%Y
  stop
 end if
end do


  


!{ Acrescenta 3 descritores relativos aos indice de confiabilidade:
!  
!  2 22 000 - Indica que as informacoes seguintes reference aos
!             dados anteriores
!  1 01 Ndesc - Indica o Numero de vezes que o proximo descritor
!             sera replicado
!  0 33 007   _ Descritor do Indice de confiabilidade

  
  ib=ib+1;b(ib)=2;a(ib)=2
  ib=ib+1;b(ib)=6;a(ib)=22
  ib=ib+1;b(ib)=8;a(ib)=000

  ib=ib+1;b(ib)=2;a(ib)=1
  ib=ib+1;b(ib)=6;a(ib)=01
  ib=ib+1;b(ib)=8;a(ib)=Ndesc


   ib=ib+1;b(ib)=2;a(ib)=0
   ib=ib+1;b(ib)=6;a(ib)=33
   ib=ib+1;b(ib)=8;a(ib)=007

 !}

  
 ! Organiza os dados da secao 3 em octetos (sec3)
 ! Neste caso o numero de octeros noct tem que ser sempre 52
 !{

   call PUT_OCTETS(A,B,ib,sec3,noct,err)  
    ! Se o número de octetos não for par
 ! acrescentar um octeto com zero
  !	{
     if (mod(noct,2)>0) then 
	  noct=noct	+1
	  sec3(noct)=char(0)
     end if
    !}

   if ((err/=0).or.(noct/=Tam_sec3)) then 
     print *,"Erro no tamanho da secao 3 : numero de octetos =",NOCT
	 stop
   end if
 !}


   

 xx=ichar(sec3(7))
 !print *,xx
 !} Fim da preparacao para gravacao da secao 3
  
  

 !{ Gravar 

  do i=1,noct
   currentRG=currentRG+1
   write (uni,rec=currentRG) sec3(i)			 
  end do




!}
!} Fim da Gravacao da secao 3


END SUBROUTINE SAVESEC3























!***********************
!*   GRAVA A SECAO 4   *
!* *********************


SUBROUTINE SAVESEC4(UN,D,ndesc,v,nsubset,iconf,tam_sec4)

 type(descbufr),dimension(:),intent(in)::D ! Descritores BUFR
 integer, intent(in)::UN				   !
 integer,intent(in)::ndesc                 ! Numero de descritores
 real,dimension(:,:),intent(in)::v
 integer,intent(out)::tam_sec4
 integer,intent(in)::nsubset
 integer,dimension(:,:),intent(in):: iconf  ! Indice de confiabilidade

 INTEGER,DIMENSION((nsubset*ndesc+2)*2)::A,B ! Vetor de Valores e Numero de bits de cada valor 
 character(len=1),dimension((nsubset*ndesc+2)*16):: sec4 ! Valor anterior e 28 
 character(len=1),dimension(3)::auxsec4          
 integer :: uni	,k,err,noct,noctaux ,j  ,i,dimab,dimoct,aa


 uni=un
 SUBNAME="SAVESEC4"
 dimab=(nsubset*ndesc+2)*2	  ! Calculo da dimensao de B() e A()
 dimoct=dimab*8				  ! Estimativa do número de octetos

 b(1)=24;a(1)=0     	  ! Tamanho da secao 4 Ainda não e conhecido 
 b(2)=8; a(2)=0           ! byte reservado (= 0)
 k=2

! Prepara para gravar todos os subsets (Valors e Indices de confiabilidade) 
! { 
 do i=1,nsubset
  !Valores
  !{
  do j=1, ndesc
    k=k+1
	b(k)=tabb(d(j)%f,d(j)%x,d(j)%y)%nbits
	a(k)=v(i,j)*10**tabb(d(j)%f,d(j)%x,d(j)%y)%scale-tabb(d(j)%f,d(j)%x,d(j)%y)%refv
  end do
  !}

  !{Indices de Confiabilidade
  do j=1,ndesc
    k=k+1
	b(k)=7   ! Número de bits do Indice de Confiabilidade
	a(k)=iconf(i,j)
	aa=a(k)
	aa=a(k)
  end do
  !}
end do


 ! Organiza os dados da secao 4 em octetos (sec4)
 ! Neste caso o numero de octeros noct tem que ser sempre 52
 !{

   call PUT_OCTETS(A,B,k,sec4,noct,err)  



 ! Se o número de octetos (Secao3+secao4)não for par
 ! acrescentar um octeto com zero
  !	{
     if (mod(noct,2)>0) then 
	  noct=noct	+1
	  sec4(noct)=char(0)
     end if
    !}


 !}
	 
  ! Agora coloca o tamanho da secao 4

   	Tam_Sec4=noct

    B(1)=24;a(1)=Tam_Sec4
    call PUT_OCTETS(A,B,1,auxsec4,noctaux,err)
	if ((noctaux/=3).or.(err/=0)) then 
	 print *,"erro secao 4 "
	 stop
	end if

	sec4(3)=auxsec4(3)
	sec4(2)=auxsec4(2)
	sec4(1)=auxsec4(1)




 !} Fim da preparacao para gravacao da secao 4

 !{ Gravar 

  do i=1,noct
   currentRG=currentRG+1
   write (uni,rec=currentRG) sec4(i)			 
  end do
!}
!} Fim da Gravacao da secao 4

  

END SUBROUTINE SAVESEC4





















 SUBROUTINE SAVESEC5(UN)
!*****************************************
!*		    GRAVAR A SECAO 5			 *
!*										 *
!*****************************************
 integer, intent(in)::UN
 integer :: uni


 integer :: i
 character(len=1),dimension(4)::sec5
 sec5(1:4) = '7'
   uni=un
 do i=1,4
  currentRg=currentRG+1
  write (uni,rec=currentRG) sec5(i)
 end do
END SUBROUTINE SAVESEC5

























! INITI_TABB
! 	Esta é uma subroutina privada para Inicializar as tabela B deste módulo
!   A subroutina carrega ler o  arquivo BufrTableB_mbufr.txt que corresponde
!   a tabela B do BUFR, em formato texto, similar ao utilizado pelo 
!   NCEP
!     Os dados desta tabela são carregados na estrutura stntab deste módulo
!     e corresponde os multiplicadores de escala , Valores de Referência e 
!     Número de Bits que são utilizados na leitura/gravação das seções 3 e 4
!     do BUFR

 SUBROUTINE INIT_TABB(Un)

  integer,intent(in)::Un
  integer::uni ,i
  integer::F,X,Y,SCALE,REFV,NBITS
  character(len=255)::C4,C5
  character(len=255)::A
  uni=un
  i=0
  	   TABB(:,:,:)%scale=0
	   TABB(:,:,:)%refv=0
	   TABB(:,:,:)%nbits=0

      OPEN (UNI, FILE ="BufrTableB_mbufr.txt", ACCESS = 'SEQUENTIAL', STATUS = 'OLD')
  
  10  READ(UNI,"(A)",END=999)A
      if (len_trim(a)>10) then  
      READ(A,100)F,X,Y,C4,C5,SCALE,REFV,NBITS
  100 FORMAT(1X,I1,I2,I3,1X,A64,1X,A17,1X,I10,1X,I12,1X,I3)
      
	  if ((F==0).and.(x<=35).and.(y<=200)) then 
		if (NBITS>256) then
		  print *,"Erro na leitura da tabela bufr"
		  print *,"Linha=",I,"Nbits=",NBITS
		  close	(uni)
		  stop
		end if

	    TABB(F,X,Y)%scale=SCALE
	    TABB(F,X,Y)%refv=REFV
	    TABB(F,X,Y)%nbits=NBITS
	  end if

	  end if
	  i=i+1
	  if (i<500) GOTO 10
999   CLOSE(UNI)

      
	  END SUBROUTINE 









	   	           

END MODULE
