
!_______________________________________________________________________
! MREADTBL - MODULO DE LEITURA DE TABELAS TEXTOS DE DADOS METEOROLOGICOS   
!            
!            Este módulo fornece suporte genérico para leitura de tabelas texto
!            compostas por colunas numerica de valores numericos
!
!            Apresenta os seguintes metedos
!
!            conf_mreadtbl = Chamada sempre antes de qualquer leitura, para
!                            informar as carateristicas da tabela a ser lida
!                            são estas
!           
!                            Totalc : Numero de colunas numericas da tabela
!                            nselc  : Numero de colunas que deseja selecionar 
!                                     para leitura. Se nsel= Totalc indica
!                                     que todas as colunas seram lidas
!                           selc(:) : Seleciona quais colunas devem ser lidas
!                                     e quais seram descartadas. Tambem indica
!                                     a ordem de leitura. Exemplo
!                         
!                                     selc(1)=2
!                                     selc(2)=1
!                                     
!                                     indica que os valores da primeira coluna !     
!                                     databela serao lidos como estivessem na 
!               !                     segunda e vice-versa
!
!            readtbl : Procede a leitura de um arquivo de tabela texto. 
!                      Durante esta leitura somente as linhas que contiverem
!                      um numero totalc de colunas  sao consideradas. 
!
!                      São considerados separadores de coluna os caracteres 
!                      branco, ! virgula, e ponto e virgula
!                      Os dados lidos são retornados na matriz obs(:,:), cujas colunas
!                      sao organizadas conforme a configuracao de selc(:) 
!
!                      No caso de existir valores  não numericos 
!                      estes são retornados como  Null=-340282300 (real ou interiro)   
!______________
MODULE MREADTBL

 USE STRINGFLIB
 USE DATELIB
  implicit none
  
PRIVATE	
 

PUBLIC conf_mreadtbl
PUBLIC Readtbl



 


 !{ Declaracao das variaveis globais
  integer :: nscols                      ! Numero de colunas selecionadas da tabela
  integer :: ncols                       ! Numero de colunas da tabela
  integer,dimension(50)::tblscol          ! numero da coluna selecionada
  integer::nbufrtbl
 !}


CONTAINS 

!_______________________________________
! Subrotine de configuracao deste modulo 
!_______________________________________
 subroutine conf_mreadtbl(totalc,nselc,selc)
    integer, intent(in):: totalc          ! Numero de colunas das tabelas
    integer, intent(in):: nselc           ! Numero de colunas selecionadas p/ leitura
    integer,dimension(:),intent(in)::selc ! Martriz com os numeros das colunas p/ leitura


!{ Declaracao de variaveis locais
    integer :: i
!}
!{ Configurar         
    nscols=nselc
    ncols=totalc

   do i=1,nselc
     tblscol(i)=selc(i)
   end do

end subroutine conf_mreadtbl 

!________________________________________________
! Subrotina de leitura de  tabela de dados Texto

 subroutine Readtbl(UN,filename,Obs,NObs)
   	integer,intent(in)::UN                  !  Unidade de Leitura
   	character(len=*),intent(in)::filename   ! Nome do arquivo de leitura
   	real,dimension(:,:),intent(inout):: Obs ! Matriz de Observacoes
   	integer,intent(out):: NObs              ! Numero de Observacoes
  
        


!{ Variaveis Locais
        character(len=60),dimension(50)::cols
        integer :: j,i
   	character(len=255):: linha
        integer:: uni
	integer::max_Nobs
!}

!{ Inicializacao de Variaveis
          uni=un 
          nobs=0
          i=0
          cols(1:50)=''
	  max_Nobs=UBOUND(obs,1)
!}
!{ Processamento - Atribui a matriz obs somente as colunas selecionadas

      
    	open(uni,file=filename,status='unknown')
        print *," "
        print *,"Lendo ",trim(filename)
 10 	read(uni,'(a)',end=999)linha
        call sep_words(linha,cols,j)
    
    	if (ncols==j) then        
         nobs=nobs+1
         do i=1,nscols 
             obs(nobs,i)=val(cols(tblscol(i)))
         end do

        end if           
    	if (nobs<max_Nobs) then 
	    goto 10
	else 
	   
	   print *,"*** Erro de Leitura *** Arquivo maior do que o previsto na matriz obs"
	end if
 999    close(uni)
 !}  
        
End subroutine readtbl



!======================================================
END MODULE
 


